// Planar shadows using the stencil buffer.
// Similar to MacReynold's old demo but rewritten for modern OpenGL.
// Includes two different, equivalent methods to create the shadow
// projection matrix.
// By Ingemar Ragnemalm 2013
// Updated 2022

// Partially based on code from Fabien Sanglard's shadow mapping
// demos, but very little remains of that.

#define MAIN
#include "MicroGlut.h"
#include <stdio.h>
#include <stdlib.h>
#include <math.h>

#include "LoadTGA.h"
#include "VectorUtils4.h"
#include "GL_utilities.h"
#include "LittleOBJLoader.h"
	// uses framework Cocoa
	// uses framework OpenGL

// ------- Application code --------

// Expressed as float so perspective division returns a float and not 0 (640/480 != 640.0/480.0).
#define RENDER_WIDTH 640.0
#define RENDER_HEIGHT 480.0

// We assign one texture unit in which to store the transformation.
#define TEX_UNIT 0

//Camera position
vec3 p_camera = vec3(52,20,0);

//Camera lookAt
vec3 l_camera = vec3(0,0,0);

//Light mouvement circle radius
float light_mvnt = 20.0f;
//float light_mvnt = 5.0f;

//Light position
vec3 p_light = vec3(0,15,0);

//Light lookAt
vec3 l_light = vec3(0,4,-10);

//vec3 potPos = {0, 4, -16};
vec3 potPos = vec3(0, 3, 0);

// Use to activate/disable projTexShader
GLuint texturedShaderId;
GLuint shadowShaderId;

mat4 AxisToZ(vec3 axis)
{
	vec3 x, y, z;
	mat4 m, R, Rt, theOp;

// Kolla ocks om parallell med Z-axel!
	if (axis.x < 0.0000001) // Under ngon tillrckligt liten grns
	if (axis.x > -0.0000001)
	if (axis.y < 0.0000001)
	if (axis.y > -0.0000001)
			return IdentityMatrix();
	
	z = Normalize(axis); // |z|
	y = Normalize(CrossProduct(z, vec3(0,0,1))); // y' = z' x z^
	x = CrossProduct(y, z); // x' = y x z
	
	R.m[0] = x.x; R.m[1] = x.y; R.m[2]  = x.z; R.m[3] = 0;
	R.m[4] = y.x; R.m[5] = y.y; R.m[6]  = y.z; R.m[13] = 0;
	R.m[8] = z.x; R.m[9] = z.y; R.m[10] = z.z; R.m[11] = 0;
	R.m[12] =   0; R.m[13] =   0; R.m[14] =   0; R.m[15] = 1;
	
	return R;
}


//-----------------------------matrices------------------------------
mat4	modelViewMatrix, textureMatrix;
mat4 projectionMatrix;
GLuint tex0, tex1; // Texture references

void loadShader()
{
	texturedShaderId = loadShaders("tex.vert", "tex.frag");
	shadowShaderId = loadShaders("shade.vert", "shade.frag");
}

// This update only change the position of the light.
void updatePositions(void)
{
	p_light.x = light_mvnt * cos(glutGet(GLUT_ELAPSED_TIME)/1000.0);
	p_light.z = light_mvnt * sin(glutGet(GLUT_ELAPSED_TIME)/1000.0);
	
	potPos.y = 3.0 + 3.0 * sin(glutGet(GLUT_ELAPSED_TIME)/500.0);
}

// Arrays for ground and the Model references

vec3 ground[] = {
							vec3(-25, 0,-25),
							vec3(-25, 0, 25),
							vec3( 25, 0, 25),
							vec3( 25, 0,-25)
							};
vec2 groundtex[] = {
							vec2(-1,-1),
							vec2(-1, 1),
							vec2( 1, 1),
							vec2( 1,-1)
							};
GLuint groundIndices[] = {0, 1, 2, 0, 2, 3};

Model *groundModel, *teapotModel;

void loadObjects(void)
{
	groundModel = LoadDataToModel(
			ground,
			NULL,
			groundtex,
			NULL,
			groundIndices,
			4,
			6);

	teapotModel = LoadModel("teapot.obj");
}

// Outer product solution from MacReynolds' demo
mat4 MakeReynoldsShadowMatrix(vec3 n, float d, vec3 l)
{
	vec4 groundplane;
	vec4 lightpos;
	groundplane = vec4(n, d);
	lightpos = vec4(l, 1);
	mat4 shadowMat;
	
	// Find the dot product between the light position vector and the ground plane normal
	float dotp = dot(groundplane, lightpos);

	shadowMat.m[0] = dotp - lightpos.x * groundplane.x;
	shadowMat.m[4] = 0.f - lightpos.x * groundplane.y;
	shadowMat.m[8] = 0.f - lightpos.x * groundplane.z;
	shadowMat.m[12] = 0.f - lightpos.x * groundplane.w;

	shadowMat.m[1] = 0.f - lightpos.y * groundplane.x;
	shadowMat.m[5] = dotp - lightpos.y * groundplane.y;
	shadowMat.m[9] = 0.f - lightpos.y * groundplane.z;
	shadowMat.m[13] = 0.f - lightpos.y * groundplane.w;

	shadowMat.m[2] = 0.f - lightpos.z * groundplane.x;
	shadowMat.m[6] = 0.f - lightpos.z * groundplane.y;
	shadowMat.m[10] = dotp - lightpos.z * groundplane.z;
	shadowMat.m[14] = 0.f - lightpos.z * groundplane.w;

	shadowMat.m[3] = 0.f - lightpos.w * groundplane.x;
	shadowMat.m[7] = 0.f - lightpos.w * groundplane.y;
	shadowMat.m[11] = 0.f - lightpos.w * groundplane.z;
	shadowMat.m[15] = dotp - lightpos.w * groundplane.w;

	return transpose(shadowMat);
}

mat4 CreateShadowProjectionMatrix(float f)
{
	mat4 m = mat4(	f, 0.0f, 0.0f, 0.0f,
						0.0f, f, 0.0f, 0.0f,
						0.0f, 0.0f, 0.0f, 0.0f,
						0.0f, 0.0f, -1.0f, f );
	return m;
}

// *** Method 1: ***
// Sequence of transformation, doing projection over an arbitrary axis.
// This is similar to the approaches in Polygons Feel No Pain for
// rotation around arbitrary axis.
	
//	1) Translate lp to origin: T(-lp)
//	2) Rotate to align with Z
//	3) Project to xy plane
//	4) Rotate back
//	5) Translate back
mat4 MakeShadowMatrix(vec3 n, float d, vec3 l)
{
	float ndotlplusd = dot(n, l) + d;
	vec3 lp = l - n * ndotlplusd; // lp = l - n(n dot l + d)
	mat4 toz = AxisToZ(n);
	mat4 fromz = transpose(toz);
	
	float f = ndotlplusd; // Distance light source to plane = focal distance
	mat4 shadowProjectionMatrix = CreateShadowProjectionMatrix(f);
	mat4 sm = T(lp.x, lp.y, lp.z) * fromz * shadowProjectionMatrix * toz * T(-lp.x, -lp.y, -lp.z);
	return sm;
}

void renderScene()
{
	mat4 mv2, tx2, trans;

	// Change light positions
	updatePositions();
	
	// Setup projection matrix
	projectionMatrix = perspective(45, RENDER_WIDTH/RENDER_HEIGHT, 1, 4000);
	// Setup the modelview from the light source
	modelViewMatrix = lookAt(p_camera.x, p_camera.y, p_camera.z,
											l_camera.x, l_camera.y, l_camera.z, 0,1,0);

	glViewport(0,0,RENDER_WIDTH,RENDER_HEIGHT);
	
	glUseProgram(texturedShaderId);
	glUniformMatrix4fv(glGetUniformLocation(texturedShaderId, "projectionMatrix"), 1, GL_TRUE, projectionMatrix.m);
	
	// Start of interesting stuff: Clear stencil, then set up to draw in it.
	glClearColor(0.2,0.2,0.5,1.0f); // Not black - we want to see if the shadow is drawn outside the plane!
	glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT | GL_STENCIL_BUFFER_BIT); // Clear Screen, Depth Buffer and Stencil Buffer
	glClearStencil(2); // No, stencil must be set to 2 outside the plane or it will draw double shadows there!

	// ** Ground in stencil and image
	glEnable(GL_STENCIL_TEST);
	glStencilFunc(GL_ALWAYS, 1, 0xFFFFFFFF);
	glStencilOp(GL_KEEP, GL_KEEP, GL_REPLACE); // Write 1 if pass! Otherwise keep! Don't draw in invisible areas.
	
	glBindTexture(GL_TEXTURE_2D, tex0);
	glUniformMatrix4fv(glGetUniformLocation(texturedShaderId, "modelViewMatrix"), 1, GL_TRUE, modelViewMatrix.m);
	DrawModel(groundModel, texturedShaderId, "in_Position", NULL, "in_Texcoord");
	
	// ** One teapot
	glDisable(GL_STENCIL_TEST);
	glBindTexture(GL_TEXTURE_2D, tex1);
	trans = T(potPos.x, potPos.y, potPos.z);
	mv2 = modelViewMatrix * trans;
	glUniformMatrix4fv(glGetUniformLocation(texturedShaderId, "modelViewMatrix"), 1, GL_TRUE, mv2.m);
	DrawModel(teapotModel, texturedShaderId, "in_Position", NULL, "in_Texcoord");
	
	// ** Teapot flattened and projected = shadow
	
	glEnable(GL_STENCIL_TEST); // Try disabling this!

	// Dumb stencil, gets problems with complex or multiple shapes. (Try it!)
	glStencilOp(GL_KEEP, GL_KEEP, GL_KEEP);

	// Use the stencil buffer to draw only once! Much better!
	glStencilFunc(GL_EQUAL, 1, 0xFFFFFFFF);
	glStencilOp(GL_KEEP, GL_KEEP, GL_INCR); // Skriv 1 om pass! Annars keep! Vi vill inte rita i osynlig yta!
	
// ** Calculate shadow matrix and draw teapot projected into plane
// Form variables vital for creating the matrix.
	
	vec3 l = p_light;
	vec3 n = vec3(0,1,0); // Det vore vldigt snyggt om jag gr planet vridbart!
	vec3 p = ground[0];
	float d = -dot(p, n);
	mat4 sm; // Shadow matrix

// *** Method 1: ***
// Sequence of transformations, doing projection over an arbitrary axis.
// This is similar to the approaches in Polygons Feel No Pain for
// rotation around arbitrary axis.

	sm = MakeShadowMatrix(n, -d, l);
	
// *** Method 2: ***
// Outer product, from Reynolds example code.

//	sm = MakeReynoldsShadowMatrix(n, -d, l);

// Use either of the matrixes (sm) above.
// They should be identical.
	
	sm = sm * T(potPos.x, potPos.y, potPos.z); // Place teapot before projecting (that is to the right)
	sm = T(0,0.001,0) * sm; // Minimal offset to be above surface. (Alternative: glDepthFunc(GL_LEQUAL) but that didn't help me here.) This is placed after projection = left
	
	glDepthFunc(GL_LEQUAL); // VIKTIG - Varfr rcker den inte?
	glBindTexture(GL_TEXTURE_2D, 0);
	mv2 = modelViewMatrix * sm;
	glUseProgram(shadowShaderId);
	glUniformMatrix4fv(glGetUniformLocation(shadowShaderId, "projectionMatrix"), 1, GL_TRUE, projectionMatrix.m);
	glUniformMatrix4fv(glGetUniformLocation(shadowShaderId, "modelViewMatrix"), 1, GL_TRUE, mv2.m);
	glUniform1f(glGetUniformLocation(shadowShaderId, "value"), 0.5); // Shadow value
	DrawModel(teapotModel, shadowShaderId, "in_Position", NULL, NULL);

	glutSwapBuffers();
}

void processNormalKeys(unsigned char key, int x, int y)
{
	if (key == 27) 
		exit(0);
}

int main(int argc, char** argv)
{
	glutInit(&argc, argv);
	glutInitDisplayMode(GLUT_RGBA | GLUT_DOUBLE | GLUT_DEPTH | GLUT_STENCIL);
	glutInitWindowPosition(100,100);
	glutInitWindowSize(RENDER_WIDTH,RENDER_HEIGHT);
	glutInitContextVersion(3, 2);
	glutCreateWindow("Planar shadow demo");
	
	loadShader();
	loadObjects();
	glEnable(GL_BLEND);
	glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
	glDisable(GL_CULL_FACE);

	LoadTGATextureSimple("floor.tga", &tex0);
	LoadTGATextureSimple("rust.tga", &tex1);
	
	glEnable(GL_DEPTH_TEST);
	glDisable(GL_CULL_FACE); // Avoid problems with missing inside.
	
	glutDisplayFunc(renderScene);
	glutRepeatingTimer(20); // MicroGlut only
	glutKeyboardFunc(processNormalKeys);
	
	glutMainLoop();
	exit(0);
}

